package com.vseravno.solna.parser;

import com.vseravno.solna.SolnaException;
import com.vseravno.solna.handler.CompiledHandler;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.NotThreadSafe;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

@NotThreadSafe
public class XmlParser {
    private static final XMLInputFactory inputFactory = XMLInputFactory.newInstance();

    private final List<CompiledHandler> handlers = new ArrayList<CompiledHandler>();

    private final InputStream stream;

    private final LinkedList<String> elementNameStack = new LinkedList<String>();

    private final LinkedList<List<XMLEvent>> aggregatedEvents = new LinkedList<List<XMLEvent>>();

    public XmlParser(@Nonnull InputStream stream, @Nonnull Collection<CompiledHandler> handlers) {
        this.stream = stream;
        this.handlers.addAll(handlers);
    }

    public void parse() throws XMLStreamException {
        XMLEventReader reader = inputFactory.createXMLEventReader(stream);

        XMLEvent startDocument = null;

        while (reader.hasNext()) {
            XMLEvent event = reader.nextEvent();

            if (event.isStartDocument()) {
                startDocument = event;
            }

            if (event.isStartElement()) {
                elementNameStack.addLast(event.asStartElement().getName().getLocalPart());
            }

            if (event.isStartElement()) {
                CompiledHandler handler = getCurrentElementHandler();
                if (handler != null) {
                    aggregatedEvents.addLast(new ArrayList<XMLEvent>());
                }
            }

            if (!aggregatedEvents.isEmpty()) {
                aggregatedEvents.getLast().add(event);
            }

            if (event.isEndElement()) {
                CompiledHandler handler = getCurrentElementHandler();
                if (handler != null) {
                    List<XMLEvent> events = aggregatedEvents.pollLast();

                    List<XMLEvent> eventsToBuild = new ArrayList<XMLEvent>();
                    eventsToBuild.add(startDocument);
                    eventsToBuild.addAll(events);
                    eventsToBuild.add(new EndDocumentImpl());

                    Object element = buildElement(handler, eventsToBuild, reader);

                    try {
                        handler.getMethodPointer().invoke(element);
                    } catch (InvocationTargetException e) {
                        throw new SolnaException("Handler thrown an exception", e);
                    }
                }
            }

            if (event.isEndElement()) {
                elementNameStack.removeLast();
            }
        }
    }

    @Nonnull
    private Object buildElement(@Nonnull CompiledHandler handler,
                                @Nonnull List<XMLEvent> events,
                                @Nonnull XMLEventReader reader) {
        try {
            ListXmlEventReader xmlEventReader = new ListXmlEventReader(events, reader);

            return handler.getAdapter().buildElement(xmlEventReader);

        } catch (Exception e) {
            throw new SolnaException("Couldn't build an element", e);
        }
    }

    @Nullable
    private CompiledHandler getCurrentElementHandler() {
        for (CompiledHandler handler : handlers) {
            StringBuilder builder = new StringBuilder();
            for (String elementName : elementNameStack) {
                builder.append("/");
                builder.append(elementName);
            }
            String currentPath = builder.toString();

            if (handler.getElementPath().equals(currentPath)) {
                return handler;
            }
        }

        return null;
    }
}
